# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("webauthn.dll")
prototypes = \
    {
        #
        'WebAuthNGetApiVersionNumber': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'WebAuthNIsUserVerifyingPlatformAuthenticatorAvailable': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pbIsUserVerifyingPlatformAuthenticatorAvailable"]),
        #
        'WebAuthNAuthenticatorMakeCredential': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_RP_ENTITY_INFORMATION", SimStruct), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_USER_ENTITY_INFORMATION", SimStruct), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_COSE_CREDENTIAL_PARAMETERS", SimStruct), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_CLIENT_DATA", SimStruct), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_AUTHENTICATOR_MAKE_CREDENTIAL_OPTIONS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("WEBAUTHN_CREDENTIAL_ATTESTATION", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWnd", "pRpInformation", "pUserInformation", "pPubKeyCredParams", "pWebAuthNClientData", "pWebAuthNMakeCredentialOptions", "ppWebAuthNCredentialAttestation"]),
        #
        'WebAuthNAuthenticatorGetAssertion': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_CLIENT_DATA", SimStruct), offset=0), SimTypePointer(SimTypeRef("WEBAUTHN_AUTHENTICATOR_GET_ASSERTION_OPTIONS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("WEBAUTHN_ASSERTION", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWnd", "pwszRpId", "pWebAuthNClientData", "pWebAuthNGetAssertionOptions", "ppWebAuthNAssertion"]),
        #
        'WebAuthNFreeCredentialAttestation': SimTypeFunction([SimTypePointer(SimTypeRef("WEBAUTHN_CREDENTIAL_ATTESTATION", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["pWebAuthNCredentialAttestation"]),
        #
        'WebAuthNFreeAssertion': SimTypeFunction([SimTypePointer(SimTypeRef("WEBAUTHN_ASSERTION", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["pWebAuthNAssertion"]),
        #
        'WebAuthNGetCancellationId': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pCancellationId"]),
        #
        'WebAuthNCancelCurrentOperation': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pCancellationId"]),
        #
        'WebAuthNGetPlatformCredentialList': SimTypeFunction([SimTypePointer(SimTypeRef("WEBAUTHN_GET_CREDENTIALS_OPTIONS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("WEBAUTHN_CREDENTIAL_DETAILS_LIST", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pGetCredentialsOptions", "ppCredentialDetailsList"]),
        #
        'WebAuthNFreePlatformCredentialList': SimTypeFunction([SimTypePointer(SimTypeRef("WEBAUTHN_CREDENTIAL_DETAILS_LIST", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["pCredentialDetailsList"]),
        #
        'WebAuthNDeletePlatformCredential': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbCredentialId", "pbCredentialId"]),
        #
        'WebAuthNGetErrorName': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeChar(label="Char"), offset=0), arg_names=["hr"]),
        #
        'WebAuthNGetW3CExceptionDOMError': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hr"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations

import itertools

from .base import SimConcretizationStrategy


class SimConcretizationStrategyNorepeats(SimConcretizationStrategy):
    """
    Concretization strategy that resolves addresses, without repeating.
    """

    def __init__(self, repeat_expr, repeat_constraints=None, **kwargs):
        super().__init__(**kwargs)
        self._repeat_constraints = [] if repeat_constraints is None else repeat_constraints
        self._repeat_expr = repeat_expr

    def _concretize(self, memory, addr, **kwargs):
        child_constraints = (*tuple(self._repeat_constraints), addr == self._repeat_expr)
        extra_constraints = kwargs.pop("extra_constraints", None)
        if extra_constraints is not None:
            child_constraints += tuple(extra_constraints)
        c = self._any(memory, addr, extra_constraints=child_constraints, **kwargs)
        self._repeat_constraints.append(self._repeat_expr != c)
        return [c]

    def copy(self):
        return SimConcretizationStrategyNorepeats(
            repeat_expr=self._repeat_expr, repeat_constraints=list(self._repeat_constraints), exact=self._exact
        )

    def merge(self, others):
        seen = {s.hash() for s in self._repeat_constraints}
        for c in itertools.chain.from_iterable(o._repeat_constraints for o in others):
            if c.hash() not in seen:
                seen.add(c.hash())
                self._repeat_constraints.append(c)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("dsound.dll")
prototypes = \
    {
        #
        'DirectSoundCreate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSound"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pcGuidDevice", "ppDS", "pUnkOuter"]),
        #
        'DirectSoundEnumerateA': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDSEnumCallback", "pContext"]),
        #
        'DirectSoundEnumerateW': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDSEnumCallback", "pContext"]),
        #
        'DirectSoundCaptureCreate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSoundCapture"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pcGuidDevice", "ppDSC", "pUnkOuter"]),
        #
        'DirectSoundCaptureEnumerateA': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDSEnumCallback", "pContext"]),
        #
        'DirectSoundCaptureEnumerateW': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDSEnumCallback", "pContext"]),
        #
        'DirectSoundCreate8': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSound8"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pcGuidDevice", "ppDS8", "pUnkOuter"]),
        #
        'DirectSoundCaptureCreate8': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSoundCapture"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pcGuidDevice", "ppDSC8", "pUnkOuter"]),
        #
        'DirectSoundFullDuplexCreate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeRef("DSCBUFFERDESC", SimStruct), offset=0), SimTypePointer(SimTypeRef("DSBUFFERDESC", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IDirectSoundFullDuplex"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSoundCaptureBuffer8"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectSoundBuffer8"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pcGuidCaptureDevice", "pcGuidRenderDevice", "pcDSCBufferDesc", "pcDSBufferDesc", "hWnd", "dwLevel", "ppDSFD", "ppDSCBuffer8", "ppDSBuffer8", "pUnkOuter"]),
        #
        'GetDeviceID': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pGuidSrc", "pGuidDest"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
import logging

from .base import SimSootStmt

l = logging.getLogger("angr.engines.soot.statements.identity")


class SimSootStmt_Identity(SimSootStmt):
    def _execute(self):
        dst = self._translate_value(self.stmt.left_op)
        src_expr = self._translate_expr(self.stmt.right_op)
        src_val = src_expr.expr
        l.debug("Identity %s := %s", dst, src_val)
        self.state.memory.store(dst, src_val)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations

from ailment.expression import Convert, BinaryOp, Const, Expression

from .base import PeepholeOptimizationExprBase


class ConstMullAShift(PeepholeOptimizationExprBase):
    """
    Convert expressions with right shifts into expressions with divisions.
    """

    __slots__ = ()

    NAME = "Conv(64->32, (N * a) >> M) => a / N1"
    expr_classes = (Convert, BinaryOp)

    def optimize(self, expr: Convert | BinaryOp, **kwargs):
        r = None

        if isinstance(expr, Convert):
            if expr.from_bits == 64 and expr.to_bits == 32:
                r = self.optimize_binaryop(expr.operand)

        elif isinstance(expr, BinaryOp):
            r = self.optimize_binaryop(expr)

        # keep size
        if r is not None and r.bits < expr.bits:
            r = Convert(expr.idx, r.bits, expr.bits, False, r, **expr.tags)

        return r

    def optimize_binaryop(self, expr: BinaryOp):
        if isinstance(expr, BinaryOp) and expr.op == "Shr" and isinstance(expr.operands[1], Const):
            # (N * a) >> M  ==>  a / N1
            inner = expr.operands[0]
            if isinstance(inner, BinaryOp) and inner.op in {"Mull", "Mul"} and not inner.signed:
                if isinstance(inner.operands[0], Const) and not isinstance(inner.operands[1], Const):
                    C = inner.operands[0].value
                    X = inner.operands[1]
                elif isinstance(inner.operands[1], Const) and not isinstance(inner.operands[0], Const):
                    C = inner.operands[1].value
                    X = inner.operands[0]
                else:
                    C = X = None

                if C is not None and X is not None:
                    V = expr.operands[1].value
                    ndigits = 5 if V == 32 else 6
                    divisor = self._check_divisor(pow(2, V), C, ndigits)
                    if divisor is not None:
                        new_const = Const(None, None, divisor, X.bits)
                        return BinaryOp(inner.idx, "Div", [X, new_const], inner.signed, **inner.tags)

        elif isinstance(expr, BinaryOp) and expr.op in {"Add", "Sub"}:
            expr0, expr1 = expr.operands
            if isinstance(expr1, Convert) and expr1.from_bits == 32 and expr1.to_bits == 64:
                r = self._match_case_a(expr0, expr1)
                if r is not None:
                    return r

            # with Convert in consideration
            if (
                isinstance(expr0, BinaryOp)
                and expr0.op in {"Shr", "Sar"}
                and isinstance(expr0.operands[1], Const)
                and expr0.operands[1].value == 0x3F
                and isinstance(expr1, BinaryOp)
                and expr1.op in {"Shr", "Sar"}
                and isinstance(expr1.operands[1], Const)
                and expr1.operands[1].value == 0x20
            ):
                expr0_operand = expr0.operands[0]
                expr1_operand = expr1.operands[0]
                if (
                    isinstance(expr0_operand, BinaryOp)
                    and expr0_operand.op in {"Mull", "Mul"}
                    and expr0_operand.signed
                    and isinstance(expr0_operand.operands[1], Const)
                ):
                    a0 = expr0_operand.operands[0]
                    a1 = expr1_operand.operands[0]
                elif (
                    isinstance(expr1_operand, BinaryOp)
                    and expr1_operand.op in {"Mull", "Mul"}
                    and expr1_operand.signed
                    and isinstance(expr1_operand.operands[1], Const)
                ):
                    a0 = expr1_operand.operands[0]
                    a1 = expr0_operand.operands[0]
                else:
                    a0, a1 = None, None

                if a0 is not None and a1 is not None and a0.likes(a1):
                    # (a * x >> 0x3f) +/- (a * x >> 0x20)  ==>  a / N
                    C = expr0_operand.operands[1].value
                    X = a0
                    V = 32
                    ndigits = 5 if V == 32 else 6
                    divisor = self._check_divisor(pow(2, V), C, ndigits)
                    if divisor is not None:
                        new_const = Const(None, None, divisor, X.bits)
                        return BinaryOp(
                            expr0_operand.idx,
                            "Div",
                            [X, new_const],
                            expr0_operand.signed,
                            **expr0_operand.tags,
                        )

        return None

    def _match_case_a(self, expr0: Expression, expr1: Convert) -> BinaryOp | None:
        # (
        #   (((Conv(32->64, vvar_44{reg 32}) * 0x4325c53f<64>) >>a 0x24<8>) & 0xffffffff<64>) -
        #   Conv(32->s64, (vvar_44{reg 32} >>a 0x1f<8>))
        # )

        expr1 = expr1.operand

        if (
            isinstance(expr0, BinaryOp)
            and expr0.op == "And"
            and isinstance(expr0.operands[1], Const)
            and expr0.operands[1].value == 0xFFFFFFFF
        ):
            expr0 = expr0.operands[0]
        else:
            return None

        if (
            isinstance(expr0, BinaryOp)
            and expr0.op in {"Shr", "Sar"}
            and isinstance(expr0.operands[1], Const)
            and isinstance(expr1, BinaryOp)
            and expr1.op in {"Shr", "Sar"}
            and isinstance(expr1.operands[1], Const)
        ):
            if (
                isinstance(expr0.operands[0], BinaryOp)
                and expr0.operands[0].op in {"Mull", "Mul"}
                and isinstance(expr0.operands[0].operands[1], Const)
            ):
                a0 = expr0.operands[0].operands[0]
                a1 = expr1.operands[0]
            elif (
                isinstance(expr1.operands[0], BinaryOp)
                and expr1.operands[0].op in {"Mull", "Mul"}
                and isinstance(expr1.operands[0].operands[1], Const)
            ):
                a1 = expr0.operands[0].operands[0]
                a0 = expr1.operands[0]
            else:
                a0, a1 = None, None

            # a0: Conv(32->64, vvar_44{reg 32})
            # a1: vvar_44{reg 32}
            if isinstance(a0, Convert) and a0.from_bits == a1.bits:
                a0 = a0.operand

            if a0 is not None and a1 is not None and a0.likes(a1):
                # (a * x >> M1) +/- (a >> M2)  ==>  a / N
                C = expr0.operands[0].operands[1].value
                X = a0
                V = expr0.operands[1].value
                ndigits = 5 if V == 32 else 6
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None:
                    new_const = Const(None, None, divisor, X.bits)
                    # we cannot drop the convert in this case
                    return BinaryOp(
                        expr0.operands[0].idx,
                        "Div",
                        [X, new_const],
                        expr0.operands[0].signed,
                        **expr0.operands[0].tags,
                    )

        return None

    @staticmethod
    def _check_divisor(a, b, ndigits=6):
        if b == 0:
            return None
        divisor_1 = 1 + (a // b)
        divisor_2 = int(round(a / float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None



from __future__ import annotations
from ailment.statement import Assignment
from ailment.expression import BinaryOp, Const, Tmp

from .base import PeepholeOptimizationStmtBase
from .utils import get_expr_shift_left_amount


class RolRorRewriter(PeepholeOptimizationStmtBase):
    """
    Rewrites consecutive statements into ROL (rotate shift left) or ROR (rotate shift right) statements.
    """

    __slots__ = ()

    NAME = "ROL/ROR rewriter"
    stmt_classes = (Assignment,)

    def optimize(self, stmt: Assignment, stmt_idx: int | None = None, block=None, **kwargs):
        # Rol example:
        #    61 | t304 = Shr32(t301,0x19)
        #    62 | t306 = Shl32(t301,0x07)
        #    63 | t303 = Or32(t306,t304)
        #
        # Ror example:
        #    98 | 0x140002a06 | t453 = (Conv(64->32, r9<8>) << 0x11<8>)
        #    99 | 0x140002a06 | t455 = (Conv(64->32, r9<8>) >> 0xf<8>)
        #    100 | 0x140002a06 | t452 = (t455 | t453)
        #
        # Another Ror example (fully propagated):
        #        | tXXX = ((t301 << 0x7<8>) | (t301 >> 0x19<8>))
        if not (isinstance(stmt.src, BinaryOp) and stmt.src.op == "Or"):
            return None

        op0, op1 = stmt.src.operands
        if isinstance(op0, Tmp) and isinstance(op1, Tmp):
            if stmt_idx < 2:
                return None

            # check the previous two instructions
            stmt_1 = block.statements[stmt_idx - 1]
            stmt_2 = block.statements[stmt_idx - 2]
            if not (isinstance(stmt_1, Assignment) and isinstance(stmt_1.src, BinaryOp)):
                return None
            if not (isinstance(stmt_2, Assignment) and isinstance(stmt_2.src, BinaryOp)):
                return None

            if not isinstance(stmt_1.dst, Tmp):
                return None
            if not isinstance(stmt_2.dst, Tmp):
                return None

            if {stmt_1.dst.tmp_idx, stmt_2.dst.tmp_idx} != {op0.tmp_idx, op1.tmp_idx}:
                return None

            stmt1_op0, stmt1_op1 = stmt_1.src.operands
            stmt2_op0, stmt2_op1 = stmt_2.src.operands

            if not (stmt1_op0.likes(stmt2_op0)):
                return None

            if not (isinstance(stmt1_op1, Const) and isinstance(stmt2_op1, Const)):
                return None

            if (
                stmt_1.src.op in {"Shl", "Mul"}
                and stmt_2.src.op == "Shr"
                and (shiftleft_amount := get_expr_shift_left_amount(stmt_1.src)) is not None
                and shiftleft_amount + stmt2_op1.value == stmt.dst.bits
            ):
                rol_amount = Const(None, None, shiftleft_amount, 8, **stmt1_op1.tags)
                return Assignment(
                    stmt.idx,
                    stmt.dst,
                    BinaryOp(None, "Rol", [stmt1_op0, rol_amount], False, bits=stmt.dst.bits, **stmt_1.src.tags),
                    **stmt.tags,
                )
            if (
                stmt_1.src.op == "Shr"
                and stmt_2.src.op in {"Shl", "Mul"}
                and (shiftleft_amount := get_expr_shift_left_amount(stmt_2.src)) is not None
                and stmt1_op1.value + shiftleft_amount == stmt.dst.bits
            ):
                return Assignment(
                    stmt.idx,
                    stmt.dst,
                    BinaryOp(None, "Ror", [stmt1_op0, stmt1_op1], False, bits=stmt.dst.bits, **stmt_1.src.tags),
                    **stmt.tags,
                )
        elif (
            isinstance(op0, BinaryOp)
            and isinstance(op1, BinaryOp)
            and {op0.op, op1.op} in [{"Shl", "Shr"}, {"Mul", "Shr"}]
        ):
            if not op0.operands[0].likes(op1.operands[0]):
                return None

            if not isinstance(op0.operands[1], Const) or not isinstance(op1.operands[1], Const):
                return None
            op0_v = op0.operands[1].value
            op1_v = op1.operands[1].value

            if (
                op0.op in {"Shl", "Mul"}
                and op1.op == "Shr"
                and (op0_shiftamount := get_expr_shift_left_amount(op0)) is not None
                and op0_shiftamount + op1_v == stmt.dst.bits
            ):
                shiftamount = Const(None, None, op0_shiftamount, 8, **op0.operands[1].tags)
                return Assignment(
                    stmt.idx,
                    stmt.dst,
                    BinaryOp(None, "Rol", [op0.operands[0], shiftamount], False, bits=stmt.dst.bits, **op0.tags),
                    **stmt.tags,
                )
            if (
                op0.op == "Shr"
                and op1.op in {"Shl", "Mul"}
                and (op1_shiftamount := get_expr_shift_left_amount(op1)) is not None
                and op0_v + op1_shiftamount == stmt.dst.bits
            ):
                shiftamount = op0.operands[1]
                return Assignment(
                    stmt.idx,
                    stmt.dst,
                    BinaryOp(None, "Ror", [op0.operands[0], shiftamount], False, bits=stmt.dst.bits, **op0.tags),
                    **stmt.tags,
                )

        return None



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("ksecdd.sys")
prototypes = \
    {
        #
        'SspiCreateAsyncContext': SimTypeFunction([], SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)),
        #
        'SspiFreeAsyncContext': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeBottom(label="Void"), arg_names=["Handle"]),
        #
        'SspiReinitAsyncContext': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Handle"]),
        #
        'SspiSetAsyncNotifyCallback': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Handle", "CallbackData"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Context", "Callback", "CallbackData"]),
        #
        'SspiGetAsyncCallStatus': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Handle"]),
        #
        'SspiAcquireCredentialsHandleAsyncW': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Arg", "Principal", "KeyVer", "Key", "Status"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "pszPrincipal", "pszPackage", "fCredentialUse", "pvLogonId", "pAuthData", "pGetKeyFn", "pvGetKeyArgument", "phCredential", "ptsExpiry"]),
        #
        'SspiAcquireCredentialsHandleAsyncA': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Arg", "Principal", "KeyVer", "Key", "Status"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "pszPrincipal", "pszPackage", "fCredentialUse", "pvLogonId", "pAuthData", "pGetKeyFn", "pvGetKeyArgument", "phCredential", "ptsExpiry"]),
        #
        'SspiInitializeSecurityContextAsyncW': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "phCredential", "phContext", "pszTargetName", "fContextReq", "Reserved1", "TargetDataRep", "pInput", "Reserved2", "phNewContext", "pOutput", "pfContextAttr", "ptsExpiry"]),
        #
        'SspiInitializeSecurityContextAsyncA': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "phCredential", "phContext", "pszTargetName", "fContextReq", "Reserved1", "TargetDataRep", "pInput", "Reserved2", "phNewContext", "pOutput", "pfContextAttr", "ptsExpiry"]),
        #
        'SspiAcceptSecurityContextAsync': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0), SimTypePointer(SimTypeRef("SecBufferDesc", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "phCredential", "phContext", "pInput", "fContextReq", "TargetDataRep", "phNewContext", "pOutput", "pfContextAttr", "ptsExpiry"]),
        #
        'SspiFreeCredentialsHandleAsync': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "phCredential"]),
        #
        'SspiDeleteSecurityContextAsync': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeRef("SecHandle", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AsyncContext", "phContext"]),
        #
        'SecMakeSPN': SimTypeFunction([SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["ServiceClass", "ServiceName", "InstanceName", "InstancePort", "Referrer", "Spn", "Length", "Allocate"]),
        #
        'SecMakeSPNEx': SimTypeFunction([SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["ServiceClass", "ServiceName", "InstanceName", "InstancePort", "Referrer", "TargetInfo", "Spn", "Length", "Allocate"]),
        #
        'SecMakeSPNEx2': SimTypeFunction([SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeChar(label="Byte"), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["ServiceClass", "ServiceName", "InstanceName", "InstancePort", "Referrer", "InTargetInfo", "Spn", "TotalSize", "Allocate", "IsTargetInfoMarshaled"]),
        #
        'SecLookupAccountSid': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="SID_NAME_USE"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Sid", "NameSize", "NameBuffer", "DomainSize", "DomainBuffer", "NameUse"]),
        #
        'SecLookupAccountName': SimTypeFunction([SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="SID_NAME_USE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("UNICODE_STRING", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Name", "SidSize", "Sid", "NameUse", "DomainSize", "ReferencedDomain"]),
        #
        'SecLookupWellKnownSid': SimTypeFunction([SimTypeInt(signed=False, label="WELL_KNOWN_SID_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SidType", "Sid", "SidBufferSize", "SidSize"]),
        #
        'MapSecurityError': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SecStatus"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    overload,
    Literal,
    Any,
)
from collections.abc import Iterable, Iterator
from dataclasses import dataclass

import networkx

import claripy
from cle.loader import Loader

from angr.code_location import CodeLocation, ExternalCodeLocation
from angr.knowledge_plugins.key_definitions.atoms import (
    Atom,
    MemoryLocation,
    AtomKind,
    Register,
    Tmp,
    ConstantSrc,
    GuardUse,
)
from angr.knowledge_plugins.key_definitions.definition import A, Definition, DefinitionMatchPredicate
from angr.knowledge_plugins.key_definitions.undefined import UNDEFINED
from angr.knowledge_plugins.cfg import CFGModel

if TYPE_CHECKING:
    pass


def _is_definition(node):
    return isinstance(node, Definition)


@dataclass
class FunctionCallRelationships:  # TODO this doesn't belong in this file anymore
    callsite: CodeLocation
    target: int | None
    args_defns: list[set[Definition]]
    other_input_defns: set[Definition]
    ret_defns: set[Definition]
    other_output_defns: set[Definition]


class DepGraph:
    """
    The representation of a dependency graph: a directed graph, where nodes are definitions, and edges represent uses.

    Mostly a wrapper around a <networkx.DiGraph>.
    """

    def __init__(self, graph: networkx.DiGraph[Definition] | None = None):
        """
        :param graph: A graph where nodes are definitions, and edges represent uses.
        """
        # Used for memoization of the `transitive_closure` method.
        self._transitive_closures: dict = {}

        if graph and not all(map(_is_definition, graph.nodes)):
            raise TypeError(f"In a DepGraph, nodes need to be <{Definition.__name__}>s.")

        self._graph: networkx.DiGraph[Definition] = graph if graph is not None else networkx.DiGraph()

    @property
    def graph(self) -> networkx.DiGraph[Definition]:
        return self._graph

    def add_node(self, node: Definition) -> None:
        """
        :param node: The definition to add to the definition-use graph.
        """
        self._graph.add_node(node)

    def add_edge(self, source: Definition, destination: Definition, **labels) -> None:
        """
        The edge to add to the definition-use graph. Will create nodes that are not yet present.

        :param source: The "source" definition, used by the "destination".
        :param destination: The "destination" definition, using the variable defined by "source".
        :param labels: Optional keyword arguments to represent edge labels.
        """
        self._graph.add_edge(source, destination, **labels)

    def nodes(self) -> networkx.classes.reportviews.NodeView[Definition]:
        return self._graph.nodes()

    def predecessors(self, node: Definition) -> Iterator[Definition]:
        """
        :param node: The definition to get the predecessors of.
        """
        return self._graph.predecessors(node)

    def transitive_closure(self, definition: Definition[Atom]) -> networkx.DiGraph[Definition[Atom]]:
        """
        Compute the "transitive closure" of a given definition.
        Obtained by transitively aggregating the ancestors of this definition in the graph.

        Note: Each definition is memoized to avoid any kind of recomputation across the lifetime of this object.

        :param definition:  The Definition to get transitive closure for.
        :return:            A graph of the transitive closure of the given definition.
        """

        def _transitive_closure(
            def_: Definition[Atom],
            graph: networkx.DiGraph[Definition[Atom]],
            result: networkx.DiGraph[Definition[Atom]],
            visited: set[Definition[Atom]] | None = None,
        ):
            """
            Returns a joint graph that comprises the transitive closure of all defs that `def_` depends on and the
            current graph `result`. `result` is updated.
            """
            if def_ in self._transitive_closures:
                closure = self._transitive_closures[def_]
                # merge closure into result
                result.add_edges_from(closure.edges())
                return result

            if def_ not in graph:
                return result

            predecessors = list(graph.predecessors(def_))

            result.add_node(def_)
            for pred in predecessors:
                edge_data = graph.get_edge_data(pred, def_)
                if edge_data is None:
                    result.add_edge(pred, def_)
                else:
                    result.add_edge(pred, def_, **edge_data)

            visited = visited or set()
            visited.add(def_)
            predecessors_to_visit = set(predecessors) - set(visited)

            closure = result
            for def0 in predecessors_to_visit:
                closure = _transitive_closure(def0, graph, closure, visited)

            self._transitive_closures[def_] = closure
            return closure

        return _transitive_closure(definition, self._graph, networkx.DiGraph())

    def contains_atom(self, atom: Atom) -> bool:
        return any(definition.atom == atom for definition in self.nodes())

    def add_dependencies_for_concrete_pointers_of(
        self, values: Iterable[claripy.ast.Base | int], definition: Definition, cfg: CFGModel | None, loader: Loader
    ):
        """
        When a given definition holds concrete pointers, make sure the <MemoryLocation>s they point to are present in
        the dependency graph; Adds them if necessary.

        :param values:
        :param definition: The definition which has data that can contain concrete pointers.
        :param cfg: The CFG, containing information about memory data.
        :param loader:
        """
        assert definition in self.nodes(), "The given Definition must be present in the given graph."

        known_predecessor_addresses: list[int | claripy.ast.Base] = [
            definition.atom.addr
            for definition in filter(lambda p: isinstance(p.atom, MemoryLocation), self.predecessors(definition))
        ]

        # concretize addresses where possible
        concrete_known_pred_addresses = []
        for address in known_predecessor_addresses:
            if isinstance(address, claripy.ast.Base):
                if address.concrete:
                    concrete_known_pred_addresses.append(address.concrete_value)
            else:
                concrete_known_pred_addresses.append(address)

        unknown_concrete_addresses: set[int] = set()
        for v in values:
            if isinstance(v, claripy.ast.Base) and v.concrete:
                v = v.concrete_value
            if isinstance(v, int) and v not in concrete_known_pred_addresses:
                unknown_concrete_addresses.add(v)

        for address in unknown_concrete_addresses:
            data_at_address = cfg.memory_data.get(address, None) if cfg is not None else None

            if data_at_address is None or data_at_address.sort not in ["string", "unknown"]:
                continue

            section = loader.main_object.find_section_containing(address)
            read_only = False if section is None else not section.is_writable
            code_location = CodeLocation(0, 0, info={"readonly": True}) if read_only else ExternalCodeLocation()

            def _string_and_length_from(data_at_address):
                if data_at_address.content is None:
                    return UNDEFINED, data_at_address.size
                return data_at_address.content.decode("utf-8"), data_at_address.size + 1

            _, string_length = _string_and_length_from(data_at_address)

            memory_location_definition = Definition(
                MemoryLocation(address, string_length),
                code_location,
            )

            self.graph.add_edge(memory_location_definition, definition)

    @overload
    def find_definitions(
        self,
        *,
        kind: type[A],
        **kwargs: Any,
    ) -> list[Definition[A]]: ...

    @overload
    def find_definitions(
        self,
        *,
        kind: Literal[AtomKind.REGISTER] = AtomKind.REGISTER,
        **kwargs: Any,
    ) -> list[Definition[Register]]: ...

    @overload
    def find_definitions(
        self,
        *,
        kind: Literal[AtomKind.MEMORY] = AtomKind.MEMORY,
        **kwargs: Any,
    ) -> list[Definition[MemoryLocation]]: ...

    @overload
    def find_definitions(
        self,
        *,
        kind: Literal[AtomKind.TMP] = AtomKind.TMP,
        **kwargs: Any,
    ) -> list[Definition[Tmp]]: ...

    @overload
    def find_definitions(
        self,
        *,
        kind: Literal[AtomKind.CONSTANT] = AtomKind.CONSTANT,
        **kwargs: Any,
    ) -> list[Definition[ConstantSrc]]: ...

    @overload
    def find_definitions(
        self,
        *,
        kind: Literal[AtomKind.GUARD] = AtomKind.GUARD,
        **kwargs: Any,
    ) -> list[Definition[GuardUse]]: ...

    @overload
    def find_definitions(
        self,
        *,
        reg_name: int | str = ...,
        **kwargs: Any,
    ) -> list[Definition[Register]]: ...

    @overload
    def find_definitions(self, *, stack_offset: int = ..., **kwargs: Any) -> list[Definition[MemoryLocation]]: ...

    @overload
    def find_definitions(self, *, const_val: int = ..., **kwargs: Any) -> list[Definition[ConstantSrc]]: ...

    def find_definitions(self, **kwargs) -> list[Definition]:
        """
        Filter the definitions present in the graph based on various criteria.
        Parameters can be any valid keyword args to `DefinitionMatchPredicate`
        """
        predicate = DefinitionMatchPredicate.construct(**kwargs)
        result = []
        defn: Definition
        for defn in self.nodes():
            if predicate.matches(defn):
                result.append(defn)
        return result

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: type[A],
        **kwargs: Any,
    ) -> list[Definition[A]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: Literal[AtomKind.REGISTER] = AtomKind.REGISTER,
        **kwargs: Any,
    ) -> list[Definition[Register]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: Literal[AtomKind.MEMORY] = AtomKind.MEMORY,
        **kwargs: Any,
    ) -> list[Definition[MemoryLocation]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: Literal[AtomKind.TMP] = AtomKind.TMP,
        **kwargs: Any,
    ) -> list[Definition[Tmp]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: Literal[AtomKind.CONSTANT] = AtomKind.CONSTANT,
        **kwargs: Any,
    ) -> list[Definition[ConstantSrc]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        kind: Literal[AtomKind.GUARD] = AtomKind.GUARD,
        **kwargs: Any,
    ) -> list[Definition[GuardUse]]: ...

    @overload
    def find_all_predecessors(
        self,
        starts: Definition[Atom] | Iterable[Definition[Atom]],
        *,
        reg_name: int | str = ...,
        **kwargs: Any,
    ) -> list[Definition[Register]]: ...

    @overload
    def find_all_predecessors(
        self, starts: Definition[Atom] | Iterable[Definition[Atom]], *, stack_offset: int = ..., **kwargs: Any
    ) -> list[Definition[MemoryLocation]]: ...

    @overload
    def find_all_predecessors(
        self, starts: Definition[Atom] | Iterable[Definition[Atom]], *, const_val: int = ..., **kwargs: Any
    ) -> list[Definition[ConstantSrc]]: ...

    def find_all_predecessors(self, starts, **kwargs):
        """
        Filter the ancestors of the given start node or nodes that match various criteria.
        Parameters can be any valid keyword args to `DefinitionMatchPredicate`
        """
        predicate = DefinitionMatchPredicate.construct(**kwargs)
        result = []
        queue = [starts] if isinstance(starts, Definition) else list(starts)
        seen = set(queue)
        while queue:
            thing = queue.pop()
            try:
                preds = self.graph.pred[thing]
            except KeyError:
                continue
            for pred in preds:
                if pred in seen:
                    continue
                queue.append(pred)
                seen.add(pred)
                if predicate.matches(pred):
                    result.append(pred)
        return result

    def find_all_successors(self, starts: Definition | Iterable[Definition], **kwargs) -> list[Definition]:
        """
        Filter the descendents of the given start node or nodes that match various criteria.
        Parameters can be any valid keyword args to `DefinitionMatchPredicate`
        """
        predicate = DefinitionMatchPredicate.construct(**kwargs)
        result = []
        queue = [starts] if isinstance(starts, Definition) else list(starts)
        seen = set(queue)
        while queue:
            thing = queue.pop()
            for pred in self.graph.succ[thing]:
                if pred in seen:
                    continue
                queue.append(pred)
                seen.add(pred)
                if predicate.matches(pred):
                    result.append(pred)
        return result

    def find_path(
        self, starts: Definition | Iterable[Definition], ends: Definition | Iterable[Definition], **kwargs
    ) -> tuple[Definition, ...] | None:
        """
        Find a path between the given start node or nodes and the given end node or nodes.
        All the intermediate steps in the path must match the criteria given in kwargs.
        The kwargs can be any valid parameters to `DefinitionMatchPredicate`.

        This algorithm has exponential time and space complexity. Use at your own risk.
        Want to do better? Do it yourself or use networkx and eat the cost of indirection and/or cloning.
        """
        return next(self.find_paths(starts, ends, **kwargs), None)

    def find_paths(
        self, starts: Definition | Iterable[Definition], ends: Definition | Iterable[Definition], **kwargs
    ) -> Iterator[tuple[Definition, ...]]:
        """
        Find all non-overlapping simple paths between the given start node or nodes and the given end node or nodes.
        All the intermediate steps in the path must match the criteria given in kwargs.
        The kwargs can be any valid parameters to `DefinitionMatchPredicate`.

        This algorithm has exponential time and space complexity. Use at your own risk.
        Want to do better? Do it yourself or use networkx and eat the cost of indirection and/or cloning.
        """
        predicate = DefinitionMatchPredicate.construct(**kwargs)
        ends = {ends} if isinstance(ends, Definition) else set(ends)
        queue: list[tuple[Definition, ...]] = (
            [(starts,)] if isinstance(starts, Definition) else [(start,) for start in starts]
        )
        seen: set[Definition] = {starts} if isinstance(starts, Definition) else set(starts)
        while queue:
            path = queue.pop()
            for succ in self.graph.succ[path[-1]]:
                newpath = (*path, succ)
                if succ in ends:
                    yield newpath
                elif succ in seen:
                    continue

                seen.add(succ)
                if predicate.matches(succ):
                    queue.append(newpath)
